/*
 * $Log:
 */

 /*
    ASCSEND  filename          : Send a file to the serial port
    CAPTURE  filename          : Capture serial input to a file
    CONFIG  filename           : Load another configuration file
    CONVERSE  "in"  "out"      : Wait for input, send output
    DIAL  name or number       : Dials from phone book (or directly) 
    DOWNLOAD  filename         : Download file
    HANGUP                     : Hang up the phone
    LOADPHONE  filename        : Load a new phonebook
    QUIT                       : Exit QTPI
    SEND string                : Send a string to the serial port
    TIMEOUT nn                 : timeout for WAIT or CONVERSE
    UPLOAD  filename           : Upload file
    WAIT string                : Waits for string to be received 
    SET variable               : Set an interval value
    GET variable               : Get an interval value
    ASKT prompt                : see below
    ASKN prompt                : see below
    STSMSG text                : writes 'text' to the QTPI status line
    ICONIFY                    : iconifies QTPI
    WAKE                       : wakes QTPI
    SAVECONF filename          : Save config
    EXITPARAM params           ; exit params
    SPAWN cmds                 ; spawn a command
    SUSPEND                    ; suspend
    RESUME                     ; resume
    VERSION                    ; reports QTPI version info
    STATUS                     ; reports QTPI status
  */

static char rcsid[] = "$Id";

#include <stdio.h>
#include <ctype.h>
#include <qptr.h>
#include <time.h>
#include "defs.h"
#include <pd/csrvthg.h>
#include "srv.h"

typedef struct
{
    char *cmd;
    int (*func) (int, void *, short);
}
CMDTABLE;

#define SM_OK   err7[0]
#define SM_NIMP err7[1]
#define SM_ERR  err7[2]
#define SM_NF   err7[3]
#define SM_ESC  err7[4]
#define SM_TIMO err7[5]
#define SM_BTMO err7[6]
#define SM_ZZZ  err7[7]
#define SM_BUSY err7[8]

short inxpr = 0;

static char *err7[] =
{
        "OK",
        "\007Not yet implemented",
        "\007Error",
	"\007Not found",
	"\007ESC",
        "\007Timeout",
        "\007Bad timeout",
        "\007Suspended",
        "\007Busy"
};

static long timeout = 20;
static long until = 0;
static int SrvFlg = 0;
static char *SrvPtr = NULL;

extern short SrvCmd;
extern char *IWait;
extern short backchat;
extern int MakeFileList (char *);
extern int RemSet(char *);
extern char *RemGet(char *);
extern void CloseSerial(void);

extern int SpawnCommand(char *);

static char srvtxt [80] = {"Startup"};

char *SrvStats(char *ctl,...)
{
    if(ctl)
    {
	va_list va;
	va_start (va, ctl);
	vsprintf (srvtxt, ctl, va);
	va_end (va);
    }
    return srvtxt;
}

int srvspwn (int ctx, void *msg, short mlen)
{
    int op1 = 0, op2 = 0;
    char *fn;
    int res = -15;
    char *resp;
    char estr[80];
    resp = SM_OK;

    if(pS->Pid)
    {
	op1 = mt_prior (pS->Pid, 0);
    }
    
    op2 = mt_prior (QTpid, 0);

    fn = stpblk ((char *) msg + 5);
    if (fn && *fn)
    {
        res = SpawnCommand(fn);
    }
    if(res < 0)
    {
        res = -res;
        strcpy(estr+1, os_errlist[res]);
        *estr = '\007';
        resp = estr;
    }
    else if(res > 0)
    {
        sprintf(resp, "\007Unknown error (%d)", res);
    }
    SrvRply (ctx, resp, -1);
    free (msg);
    if(op1)
	mt_prior (pS->Pid, op1);
    mt_prior (QTpid, op2);
    return 0;
}
int srvsusp (int ctx, void *msg, short mlen)
{
    short zzz;
    char *rply;
    int op1, op2;
    
    CloseSerial();
    op1 = mt_prior (ClkPID, 0);
    op2 = mt_prior (QTpid, 0);
    SrvRply (ctx, SM_OK, -1);
    xfree (msg);

    for(zzz = 1; (zzz & 1);)
    {
        if((ctx = SrvPoll ("QTPI", -1, &msg, &mlen)) >= 0)
        {
            if(strnicmp(msg, "resume", 6) == 0)
            {
                zzz = 0;
	    } 
            if(strnicmp(msg, "quit", 4) == 0)
            {
                zzz = 2;
	    }
            if(zzz & 1)
            {
                rply = SM_ZZZ;
            }
            else
            {
                rply = SM_OK;
	    } 
            SrvRply (ctx, rply, -1);
            free (msg);
        }
    } 
    mt_prior (ClkPID, op1);
    mt_prior (QTpid, op2);
    SetSerial(1);    
    if(zzz == 2)
    {
        SrvCmd = SRV_QUIT;
        pick (pChId, -1, QTpid, K_WAKE);
    }
    return 0;
}

int srvexp (int ctx, void *msg, short mlen)
{
    char *fn;
    
    fn = stpblk ((char *) msg + 10);
    if (fn && *fn)
    {
        if(ExitParams)
        {
            xfree(ExitParams);
        }
        ExitParams = mystrdup(fn);
    }
    SrvRply (ctx, SM_OK, -1);
    free (msg);
    return 0;
}
int srvtimo (int ctx, void *msg, short mlen)
{
    int to;
    char *q;
    char *rply = SM_BTMO;
    
    to = strtol (((char *) (msg) + 7), &q, 0);
    if (to != 0)
    {
	if (q)
	{
	    strupr (q);
	    if (strstr (q, "min"))
	    {
		to *= 60;
	    }
	}
	if (to >= 0)
	{
	    timeout = to;
	}
	else
	{
	    timeout = -1;
	}
        rply = SM_OK;
    }
    SrvRply (ctx, rply, -1);
    free (msg);
    return 0;
}

int srvwait (int ctx, void *msg, short mlen)
{
    char *fn;

    fn = stpblk ((char *) msg + 4);
    if (fn && *fn)
    {
	SrvFlg = (int) InitWait (fn);
	if (timeout > 0)
	{
	    until = time (NULL) + timeout;
	}
	else
	{
	    until = (long) 0xffffffff;
	}
	return 1;
    }
    else
    {
    	SrvRply (ctx, SM_ERR, -1);
	free (msg);
	return 0;
    }
}

int srvdial (int ctx, void *msg, short mlen)
{
    char *fn;

    fn = stpblk ((char *) msg + 4);
    if (fn && *fn)
    {
        extern void InitDialLine(void);
        InitDialLine();
	if (SearchBook (fn) == 0)
	{
	    strcpy (pPhoneNo, fn);
	    ReDial (NULL);
	}
	SrvFlg = Q_DIAL;
	SrvCmd = SRV_DOZE;
	pick (pChId, -1, QTpid, K_WAKE);
	return 2;
    }
    else
    {
	SrvRply (ctx, SM_ERR, -1);
	free (msg);
	return 0;
    }

}

int srvsavcfg (int ctx, void *msg, short mlen)
{
    char *rply;
    char *fn;
    fn = stpblk ((char *) msg + 8);
    if (fn && *fn)
    {
	strcpy (QeM_Dat, fn);
    }

    pSave ();
    SrvRply (ctx, SM_OK, -1);
    free (msg);
    return 0;
}

int srvset (int ctx, void *msg, short mlen)
{
    char *fn;
    int res;
    fn = stpblk ((char *) msg + 3);
    if (fn && *fn)
    {
        res = RemSet(fn);
    }
    SrvRply (ctx, (res) ? SM_OK : SM_ERR, -1);
    free (msg);
    return 0;
}

int srvget (int ctx, void *msg, short mlen)
{
    char * res, *fn, *p;
    fn = stpblk ((char *) msg + 3);
    if (fn && *fn)
    {
        res = RemGet (fn);
        if(p = strchr(res, '='))
        {
            res = ++p;
        }
    }
    SrvRply (ctx, (res) ? res : SM_ERR, -1);
    free (msg);
    return 0;
}

int srvproto (int ctx, void *msg, short mlen)
{
    char *fn;
    char *rpt;

    rpt = SM_ERR;
    fn = stpblk ((char *) msg + 8);
    if (fn && *fn)
    {
	if (FindXPRExe (fn))
	{
	    strcpy (MyProtocol, fn);
	    XPRStatus = XPRUse (fn);
	    strcpy (xproto, fn);
	    rpt = SM_OK;
	}
    }
    SrvRply (ctx, rpt, -1);
    free (msg);
    return 0;
}

int srvasc (int ctx, void *msg, short mlen)
{
    int ssts = 0;
    char *fn;
    char *rmsg;
    int rv = 0;

    fn = stpblk ((char *) msg + 7);
    if (fn == NULL)
    {
	fn = "";
    }
    ssts = DoSendItem (fn, 0);
    if (ssts > 0)
    {
	rv = 3;
	SrvFlg = pSendId;
    }
    else if (ssts == 0)
    {
	rmsg = "Cancelled";
    }
    else
    {
	rmsg = SM_NF;
    }

    if (rv == 0)
    {
	SrvRply (ctx, rmsg, -1);
	free (msg);
    }
    return rv;
}

int srvcap (int ctx, void *msg, short mlen)
{
    int ssts = 0;
    char *fn;
    char *rmsg;

    fn = stpblk ((char *) msg + 7);
    if (fn == NULL)
    {
	fn = "";
    }
    ssts = DoCaptureItem (fn, 1, 0);	/* now, non-interactive */

    if (ssts > 0)
    {
	rmsg = SM_OK;
    }
    else
    {
	rmsg = "Cancelled";
    }

    SrvRply (ctx, rmsg, -1);
    free (msg);
    return 0;
}

int srvcfg (int ctx, void *msg, short mlen)
{
    char *rply;
    char *fn;

    fn = stpblk ((char *) msg + 6);
    if (fn && *fn)
    {
	char *op = mystrdup (QeM_Dat);

	strcpy (QeM_Dat, fn);
	if (pRestore () == 0)
	{
	    strcpy (QeM_Dat, op);
	    rply = SM_ERR;
	}
	else
	{
            SetParams(0);
	    rply = SM_OK;
	}
	xfree (op);
    }
    SrvRply (ctx, rply, -1);
    free (msg);
    return 0;
}

int srvcvrs (int ctx, void *msg, short mlen)
{
    char *fn;

    fn = stpblk ((char *) msg + 8);
    if (fn && *fn)
    {
	char *p;

	p = strchr (fn, 0);
	if (p)
	{
	    if ((p - msg) < mlen)
	    {
		if (*++p)
		{
		    ParseString (fn, (int (*)(char, char **)) pS->SendChr, 1, NULL);
		    SrvFlg = (int) InitWait (p);
		    if (timeout > 0)
		    {
			until = time (NULL) + timeout;
		    }
		    else
		    {
			until = (long) 0xffffffff;
		    }
		    return 1;
		}
	    }
	}
    }
    SrvRply (ctx, SM_ERR, -1);
    free (msg);
    return 0;
}

int srvdld (int ctx, void *msg, short mlen)
{
    char *fn;

    fn = stpblk ((char *) msg + 8);
    SrvCmd |= SRV_DN;
    SrvPtr = fn;
    pick (pChId, -1, QTpid, K_WAKE);
    return SRV_DN;
}

int srvhng (int ctx, void *msg, short mlen)
{
    pS->SendChr ('\r');
    ParseString (MdmCtrl[0], (int (*)(char, char **)) pS->SendChr, 1, NULL);
    GoOffLine ();
    SrvRply (ctx, SM_OK, -1);
    free (msg);
    return 0;
}

int srvlpn (int ctx, void *msg, short mlen)
{
    char *rply;
    char *fn;

    fn = stpblk ((char *) msg + 6);
    if (fn && *fn)
    {
	char *op = mystrdup (Qbook_dat);

	strcpy (Qbook_dat, fn);
	if (BuildList (NULL) != 1)
	{
	    strcpy (Qbook_dat, op);
	    rply = SM_ERR;
	}
	else
	{
	    rply = SM_OK;
	}
	xfree (op);
    }
    SrvRply (ctx, rply, -1);
    free (msg);
    return 0;
}

int srvquit (int ctx, void *msg, short mlen)
{
    free (msg);
    SrvCmd = SRV_QUIT;
    SrvRply (ctx, SM_OK, -1);
    pick (pChId, -1, QTpid, K_WAKE);
    return 0;
}

int srvvers (int ctx, void *msg, short mlen)
{
    char rep[32];

    SrvRply (ctx, "QTPI v" QTPI_VERS, -1);
    free (msg);
    return 0;
}

int srvsend (int ctx, void *msg, short mlen)
{
    ParseString (((char *) msg + 5), (int (*)(char, char **)) pS->SendChr, 1, NULL);
    SrvRply (ctx, SM_OK, -1);
    free (msg);
    return 0;
}

int srvuld (int ctx, void *msg, short mlen)
{
    char *fn;

    fn = stpblk ((char *) msg + 6);
    if (fn && *fn)
    {
	SrvCmd |= SRV_UP;
	SrvPtr = fn;
	pick (pChId, -1, QTpid, K_WAKE);
	return SRV_UP;
    }
    SrvRply (ctx, SM_ERR, -1);
    free (msg);
    return 0;
}

int srvASK (int ctx, void *msg, short mlen, short flag)
{
    short ok = 0;
    char *fn;

    fn = stpblk ((char *) msg + 4);
    if (fn && *fn)
    {
	char *p;

	if (flag == SRV_AN)
	{
	    p = strchr (fn, 0);
	    if (p)
	    {
		if ((p - (char *) msg) < mlen)
		{
		    if (*++p)
		    {
			ok = 1;
		    }
		}
	    }
	}
	else
	{
	    ok = 1;
	}
	if (ok)
	{
	    short tlen;
	    tlen = mlen - (fn - (char *) msg);
	    SrvCmd |= flag;
	    SrvPtr = xmalloc(mlen);
            memcpy(SrvPtr, fn, tlen);
	    pick (pChId, -1, QTpid, K_WAKE);
	    return flag;
	}
    }
    SrvRply (ctx, SM_ERR, -1);
    free (msg);
    return 0;
}

int srvaskt (int ctx, void *msg, short mlen)
{
    return srvASK (ctx, msg, mlen, SRV_AT);
}

int srvaskn (int ctx, void *msg, short mlen)
{
    return srvASK (ctx, msg, mlen, SRV_AN);
}

int srvsts (int ctx, void *msg, short mlen)
{
    char *fn;

    fn = stpblk ((char *) msg + 6);
    if (fn && *fn)
    {
	OutError (fn);
    }
    SrvRply (ctx, SM_OK, -1);
    free (msg);
    return 0;
}

extern short iconic;

int srvicon (int ctx, char *msg, short mlen)
{
    if (QAwake)
    {
	if(*(msg+1) == 'S' || *(msg+1) == 's')
	{
	    iconic = 1;
	}
	SrvCmd |= SRV_SLEEP;
	pick (pChId, -1, QTpid, K_WAKE);
    }
    SrvRply (ctx, SM_OK, -1);
    free (msg);
    return 0;
}

int srvwake (int ctx, char *msg, short mlen)
{
    if(*msg == 'i' || *msg == 'I')
    {
	iconic = 0;
    }

    if(*(msg + 1) != 'F' && *(msg + 1) != 'f' )
    {
	pick (pChId, -1, QTpid, K_WAKE);
    }
    
    SrvRply (ctx, SM_OK, -1);
    free (msg);
    return 0;
}

int srvstatus (int ctx, void *msg, short mlen)
{
    if(qC->qState == Q_ONLINE)
    {
	time_t ol = time(NULL) - qC->qClock;
	SrvStats("Online %s %02d:%02d:%02d", CheckPhone(pPhoneNo),
		 ol/3600,
		 (ol/60) % 60,
		 ol % 60);
    }
    SrvRply (ctx, srvtxt, -1);
    free (msg);
    return 0;
}

int srvecho (int ctx, void *msg, short mlen)
{
    char *fn;
    char *rply = SM_OK;

    fn = stpblk ((char *) msg + 4);
    if (fn && *fn)
    {
	if (stricmp (fn, "on") == 0)
	{
	    backchat = 1;
	}
    	else if (stricmp (fn, "off") == 0)
	{
	    backchat = 0;
	}
	else
	{
	    rply = SM_ERR;
	}
    }

    SrvRply (ctx, rply, -1);
    free (msg);
    return 0;
}

static CMDTABLE pcmdtbl[] =
{
    {"TIMEOUT", srvtimo},
    {"WAIT", srvwait},
    {"DIAL", srvdial},
    {"SET", srvset},
    {"GET", srvget},
    {"PROTOCOL", srvproto},
    {"ASCSEND", srvasc},
    {"CAPTURE", srvcap},
    {"CONFIG", srvcfg},
    {"CONVERSE", srvcvrs},
    {"DOWNLOAD", srvdld},
    {"HANGUP", srvhng},
    {"LOADPHONE", srvlpn},
    {"QUIT", srvquit},
    {"SEND", srvsend},
    {"UPLOAD", srvuld},
    {"ASKT", srvaskt},
    {"ASKN", srvaskn},
    {"STSMSG", srvsts},
    {"ICONIFY", (int (*)(int , void *, short ))srvicon},
    {"ISLAVE", (int (*)(int , void *, short ))srvicon},
    {"IFREE", (int (*)(int , void *, short ))srvwake},
    {"WAKE", (int (*)(int , void *, short ))srvwake},
    {"IWAKE", (int (*)(int , void *, short ))srvwake},
    {"ECHO", srvecho},
    {"SAVECONF", srvsavcfg},
    {"EXITPARAM", srvexp},
    {"SUSPEND", srvsusp},
    {"SPAWN", srvspwn},
    {"VERSION", srvvers},
    {"STATUS", srvstatus},
    {NULL}
};

short StartServer (void)
{
    return UseSrvThg (SRVR, "QTPI");
}

short srv_in_use = 0;

extern char *ReturnDialLine (void);

int PollServer (void)
{
    static int waiting = 0;
    static void *dptr;
    static int ctx;
    char dlin[82];
    
    short dlen;

    if (srv_in_use)
    {
	char *rptr;

	switch (srv_in_use)
	{
	    case 1:
		if (SrvFlg != (int) IWait)
		{
		    rptr = ReturnLastLine ();
		    srv_in_use = 0;
		}
		else if (until != 0xffffffff)
		{
		    if (until < time (NULL))
		    {
			srv_in_use = 0;
			InitWait (NULL);
			rptr = SM_TIMO;
		    }
		}
		break;
#if 1
	    case SRV_UP:
	    case SRV_DN:
		if ((SrvCmd & srv_in_use) == 0)
		{
		    srv_in_use = 0;
		    rptr = SM_OK;
		}
		break;
#endif
	    case 3:
		if (pSendId != SrvFlg)
		{
		    srv_in_use = 0;
		    rptr = SM_OK;
		}
		break;
	    case 2:
		if (((SrvCmd & SRV_DOZE) == 0) && qC->dState == 0)
		{
		    srv_in_use = 0;
		    rptr = strcpy(dlin, ReturnDialLine ());
		}
		break;
	    case SRV_AT:
	    case SRV_AN:
		if ((SrvCmd & srv_in_use) == 0)
		{
		    srv_in_use = 0;
		    rptr = SrvPtr;
		}
		break;
	}
	if (srv_in_use == 0)
	{
	    int n,j;
	    n = strlen(rptr);
	    j = SrvRply (ctx, rptr, n);
	    free (dptr);
	}
    }
    else
    {
	ctx = SrvPoll ("QTPI", 0, &dptr, &dlen);
	if (ctx >= 0)
	{
	    if(inxpr)
	    {
		if(strnicmp(dptr, "STATUS", 6) == 0)
		{
		    srvstatus (ctx, dptr, dlen);
		}
		else
		{
		    SrvRply (ctx, SM_BUSY, -1);
		    if (dptr)
		    {
			free (dptr);
		    }
		}
	    }
	    else
	    {
		CMDTABLE *pct;

		for (pct = pcmdtbl; pct->cmd; pct++)
		{
		    if (dptr)
		    {
			if (strnicmp (pct->cmd, dptr, strlen (pct->cmd)) == 0)
			{
			    srv_in_use = (pct->func) (ctx, dptr, dlen);
			    break;
			}
		    }
		}
		if (pct->cmd == NULL)
		{
		    OutError ("Unrecognised REQUEST %s", dptr);
		    SrvRply (ctx, SM_ERR, -1);
		    if (dptr)
		    {
			free (dptr);
		    }
		}
	    }
	}
    }
}

void ServerXPR (short flag)
{
    if (flag == SRV_DN)
    {
	XPRIO (SrvPtr, RECEIVE);
    }
    else
    {
	if (MakeFileList (SrvPtr) > 0)
	{
	    XPRIO (NULL, SEND);
	}
    }
}

static char reply[64];

void ServerASK (struct WM_wwork *wwk, short flag)
{
    int res;
    char *p, *pt[6];

    p = strchr (SrvPtr, 0);
    Blocked = 1;
    *reply = 0;
    
    if (flag == SRV_AT)
    {
	if (p && *(p+1))
	{
	    strncpy (reply, p + 1, 60);
	}
	res = menu_text (wwk, SrvPtr, reply, 60);
        xfree (SrvPtr);
	if (res == -1)
	{
	    SrvPtr = SM_ESC;
	}
	else
	{
	    SrvPtr = reply;
	}
    }
    else
    {
	char *q, *r;
	int nopt = 0;

	p++;

	for (q = p; r = strtok (q, "|"); q = NULL)
	{
	    if (*r)
	    {
                *r = toupper(*r);
		pt[nopt] = r;
		nopt++;
		if (nopt == 6)
		{
		    break;
		}
	    }
	}

	res = menu_choice (wwk, SrvPtr, nopt, 0,
			   pt[0], pt[1], pt[2], pt[3], pt[4], pt[5], pt[6]);
        xfree (SrvPtr);
	sprintf (reply, "%d", res);
	SrvPtr = reply;
    }
    Blocked = 0;
}

